/*========================================================================

  File: EPPSTEIN.c

  Program EPPSTEIN - Version 1.1 - July 1999

  ========================================================================

  This module contains an implementation of Eppstein's Algorithm
  that enumerates (by increasing weight) the N shortest paths in
  weighted graphs. The algorithm is described in: 

         "Finding the K Shortest Paths" by David Eppstein,
         SIAM J. COmput. vol. 28 no. 2 pp. 652-673 (1999)

  This implementation corresponds to  what is called the basic version
  in this paper. After the  shortest paths tree has been computed, the
  K shortest  paths are enumerated  in time  O(m+ n log  n + K  log K)
  where n  is the  number of  nodes and m  the number  of arcs  in the
  graph.

  ========================================================================

    Copyright (C) 1999 Victor Jimenez and Andres Marzal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (file COPYING); if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    You can contact the authors at:

           Victor Jimenez and Andres Marzal
           Dept. de Informatica, Universitat Jaume I, Castellon, Spain
           {vjimenez,amarzal}@inf.uji.es

    Updated information about this software will be available at the
    following www address:

           http://terra.act.uji.es/REA
 
  ========================================================================= */

#define VERSION "1.1"


#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <string.h>

#ifdef DEBUG
#include <assert.h>
#endif

#include <EPPSTEIN.h>
#include <loadgraph.h>
#include <dijkstraRev.h>
#include <chronometer.h>

typedef struct DevPath {
  COST_TYPE cost;
  struct DevPath *parent;  /* Previous element of the sequence */
  struct Heap3Node *heap3; /* Last element of the sequence     */
} DevPath;    /* A path represented by a sequence of deviations */
              /* from the shortest paths tree                   */

typedef struct DevPath *PtrDevPath;


/*============================================================================
  GLOBAL VARIABLES
  ============================================================================*/

Heap3Node *heap3;
Heap3Node *firstFreeHeap3Node;
int heap3Size;

DevPath    *devPaths;
DevPath    *firstFreeDevPath;
PtrDevPath *searchHeap;
int searchHeapSize = 0;

PtrDevPath *bestDevPaths;
int numberComputedPaths = 0;

PtrNode *stack;
PtrDevPath * traceBackStack;

float *cumulatedSeconds;

/*============================================================================*/
Heap3Node * Heap3Malloc() {
  Heap3Node * oldFirst = firstFreeHeap3Node;

#ifdef DEBUG  
  if (firstFreeHeap3Node - heap3 > heap3Size - 1) {
    printf("IMPOSSIBLE ERROR: There are no free nodes at heap3\n");
    exit(1);
  }
#endif

  firstFreeHeap3Node++;
  return oldFirst;
}


/*============================================================================*/
void HeapifyOutHeap (Heap3Node *heap, int heapSize) {
  /* Heapifies from heap[1] to heap[heapSize], moving only the data field */
  
  register int i;
  for (i = heapSize/2; i >= 1; i--) {
    Heap3Data item = heap[i].data;
    register int j = 2*i;
    while (j <= heapSize) {
      if (j+1 <= heapSize && heap[j+1].data.cost < heap[j].data.cost) j += 1;
      if (item.cost <= heap[j].data.cost) break;
      heap[j/2].data = heap[j].data;
      j = 2*j;
    }
    heap[j/2].data = item;
  }
}


/*============================================================================*/
void BuildOutHeap (Node *node) {
  /* Builds a 2-heap of the edges in G-T with tails at the node,
     with the restriction that the root of the heap only has one child.
     The weight of an edge e is delta(e).
  */

  PtrArc *arc;
  int numberArcs;
  Heap3Data auxData;
  Heap3Node *bestPos, *pos, *father, *child;
  COST_TYPE bestCost;
  int heapSize = 0;
  
#ifdef DEBUG
  assert (node!=NULL);
#endif
  
  if (node->bestCost >= INFINITY_COST) {
    node->heap3 = NULL;
    node->heap3Done = 1;
    return;
  }
  
  bestCost = INFINITY_COST;
  bestPos = NULL;
  for (arc = node->firstArcOut, numberArcs = node->numberArcsOut;
       numberArcs != 0; arc++, numberArcs--) {
    if ( (*arc)->dest != node->bestPath && (*arc)->dest->bestCost < INFINITY_COST) {
      COST_TYPE delta = (*arc)->cost + (*arc)->dest->bestCost - node->bestCost;
      if (bestPos == NULL) node->heap3 = pos = bestPos = Heap3Malloc ();
      else pos = Heap3Malloc ();
      heapSize++;
      pos->data.cost = delta;
      pos->data.arc  = *arc;
      if (delta < bestCost) {
	bestCost = delta;
	bestPos  = pos;
      }
    }   
  }
  
  if (heapSize==0) {
    node->heap3 = NULL;
    node->heap3Done = 1;
  } else {
    if (bestPos != node->heap3) {
      auxData = bestPos->data;
      bestPos->data = node->heap3->data;
      node->heap3->data = auxData;
    }
    
    node->heap3[0].left = node->heap3[0].right = NULL;
    if (heapSize > 1) node->heap3[0].center = node->heap3 + 1;
    else node->heap3[0].center = NULL;
    
    for (father=node->heap3 + 1, child=node->heap3 + 2; child < node->heap3 + heapSize; father++) {
      father->left = child++;
      father->right = (child < node->heap3 + heapSize) ? child++ : NULL;
      father->center = NULL;
    }
    for ( ; father < node->heap3 + heapSize; father++)
      father->left = father->right = father->center = NULL;
    
    if (heapSize > 2)
      HeapifyOutHeap (node->heap3, heapSize-1); /* Heapify from node->heap3[1] */
  }
  
}


/*============================================================================*/
Heap3Node * MergeHeap3(Heap3Node * outHeap, Heap3Node * devHeap) {
  /* Inserts the root of outHeap in devHeap using persistence techniques */
  
  Heap3Node * newNode;
  
  if (devHeap == NULL) return outHeap;
  if (outHeap == NULL) return devHeap;
  
  if (outHeap->data.cost <= devHeap->data.cost) {
    if (devHeap->desc == 0) {
      outHeap->left = devHeap;
      outHeap->desc = 1;
      return outHeap;
    }
    
    newNode = Heap3Malloc();
    newNode->data = devHeap->data;
    newNode->left = newNode->right = NULL;
    newNode->center = devHeap->center;
    newNode->desc = 0;
    
    if (devHeap->right!=NULL && devHeap->right->desc > devHeap->left->desc) {
      outHeap->left = MergeHeap3(newNode, devHeap->left);
      outHeap->right = devHeap->right;
    }
    else {
      outHeap->left = devHeap->left;
      outHeap->right = MergeHeap3(newNode, devHeap->right);      
    }
    outHeap->desc = devHeap->desc + 1;
    return outHeap;
  }
  else {
    newNode = Heap3Malloc();
    
    *newNode = *devHeap;
    if (newNode->left == NULL ||
        (newNode->right!=NULL && newNode->right->desc > newNode->left->desc)) {
      newNode->left = MergeHeap3(outHeap, newNode->left);
    }
    else {
      newNode->right = MergeHeap3(outHeap, newNode->right);      
    }
    newNode->desc++;
    return newNode;
  }
}


/*============================================================================*/
void BuildDeviationHeaps (Graph * graph) { 

  int  numberNodes;
  Node *node, *node2;
  int top = -1;
  
  BuildOutHeap(graph->finalNode);
  graph->finalNode->heap3Done = 1;
  
  for (node = graph->node, numberNodes = graph->numberNodes;
       numberNodes != 0; node++, numberNodes--) {
    if (node->heap3Done == 1) continue;
    
    for (node2 = node; node2->heap3Done != 1 && node2->bestPath != NULL;
         node2 = node2->bestPath) 
      stack[++top] = node2;
    
    if (top > -1) {
      node2 = stack[top--];
      BuildOutHeap (node2);
      
      if (node2->bestPath != NULL) 
	node2->heap3 = MergeHeap3 (node2->heap3, node2->bestPath->heap3);
      node2->heap3Done = 1;
      
      for ( ; top >= 0; --top) {
	node2 = stack[top];
#ifdef DEBUG
	assert (node2 != NULL);
#endif
	BuildOutHeap (node2);
	node2->heap3 = MergeHeap3 (node2->heap3, stack[top+1]->heap3);
	node2->heap3Done = 1;
      }
    }
  }
}


/*============================================================================*/
void RemoveMinFromSearchHeap(void) {

  if (searchHeapSize == 0) return;
  searchHeap[0] = searchHeap[--searchHeapSize];
  {
    register DevPath * item = searchHeap[0];
    register int j = 1;
    while (j<searchHeapSize) {
      if (j<searchHeapSize-1 && searchHeap[j+1]->cost < searchHeap[j]->cost) j++;
      if (item->cost <= searchHeap[j]->cost) break;
      searchHeap[(j-1)/2] = searchHeap[j];
      j = 2*j+1;
    }
    searchHeap[(j-1)/2] = item;
  }
}


/*============================================================================*/
void InsertInSearchHeap(DevPath * cand) {

  register int child, father;
  
  child = searchHeapSize++;
  while (child > 0 && searchHeap[father=(child-1)>>1]->cost > cand->cost) {
    searchHeap[child] = searchHeap[father];
    child = father;
  }
  searchHeap[child] = cand;
  
}


/*============================================================================*/
void SearchDeviationPaths (Graph *graph, int K) {
  
  if (graph->initialNode->heap3 == NULL) return;
  bestDevPaths[0] = NULL;
  numberComputedPaths=1;
  
  searchHeap[0] = firstFreeDevPath++;
  searchHeap[0]->parent = NULL;  
  searchHeap[0]->heap3 = graph->initialNode->heap3;
  searchHeap[0]->cost = graph->initialNode->heap3->data.cost;
  searchHeapSize=1;
  
  while(numberComputedPaths < K && searchHeapSize > 0) {
    DevPath * best = bestDevPaths[numberComputedPaths] = searchHeap[0];
    Heap3Node * bestHeap3 = best->heap3;
    
    RemoveMinFromSearchHeap();
    if (bestHeap3->left != NULL) {
      DevPath * cand = firstFreeDevPath++;
      cand->parent = best->parent;
      cand->heap3  = bestHeap3->left;
      cand->cost   = best->cost + (bestHeap3->left->data.cost - bestHeap3->data.cost);
      InsertInSearchHeap(cand);
    }
    if (bestHeap3->right != NULL) {
      DevPath * cand = firstFreeDevPath++;
      cand->parent = best->parent;
      cand->heap3  = bestHeap3->right;
      cand->cost   = best->cost + (bestHeap3->right->data.cost - bestHeap3->data.cost);
      InsertInSearchHeap(cand);
    }
    if (bestHeap3->center != NULL) {
      DevPath * cand = firstFreeDevPath++;
      cand->parent   = best->parent;
      cand->heap3    = bestHeap3->center;
      cand->cost     = best->cost + (bestHeap3->center->data.cost - bestHeap3->data.cost);
      InsertInSearchHeap(cand);
    }
    
    
    if (bestHeap3->data.arc->dest->heap3 != NULL) {
      DevPath * cand = firstFreeDevPath++;
      cand->parent   = best;
      cand->heap3    = bestHeap3->data.arc->dest->heap3;
      cand->cost     = best->cost + cand->heap3->data.cost;
      InsertInSearchHeap(cand);
    }
    cumulatedSeconds[numberComputedPaths] = ClockTotal();
    numberComputedPaths++;
  }
}


/*============================================================================*/
void PrintBestPath (Node *node) { 

  Node *nextNode;
  
#ifdef DEBUG
  assert (node != NULL); 
#endif

  if (node->bestCost < INFINITY_COST) {
    for (nextNode = node; nextNode != NULL; nextNode = nextNode->bestPath)
      printf ("%i-", nextNode->name);
    printf (" \t(Cost: %7.2f)", node->bestCost);
  }
}


/*============================================================================*/
void Copyright () {
  printf ("\n=========================================================================\n");
  printf ("EPPSTEIN version %s, Copyright (C) 1999 Victor Jimenez and Andres Marzal\n",
	  VERSION);
  printf ("EPPSTEIN comes with ABSOLUTELY NO WARRANTY.\n");
  printf ("This is free software, and you are welcome to redistribute it under\n");
  printf ("certain conditions; see the README and COPYING files for more details.\n");
}


/*============================================================================*/
void Help (char *program) {
  printf ("=========================================================================\n");
  printf ("\nUse: %s GRAPH_FILE NUMBER_OF_PATHS [-paths] [-tdijkstra]\n", program);
  printf ("     Optional arguments:\n");
  printf ("       -paths Print the sequence of nodes for each path\n");
  printf ("       -tdijkstra Cumulate also the time of Dijkstra's algorithm\n");
  printf ("See the README file for more details.\n");
  printf ("=========================================================================\n");
  exit (1);
}


/*============================================================================*/
int main (int argc, char **argv) {
  FILE *outf;
  Graph  graph;
  Node   *node;
  int    i, numberPaths = 1;
  time_t date;
  struct tm *localDate;
//  char   hostName[200] = "";
  int    showPaths = 0;
  int    measureDijkstra = 0;
#ifdef DEBUG
  int    numberNodes;
#endif
  
  /****************** Prints the copyright notice  **************************/
  //Copyright ();
  
  /************** Reads the command line parameters *************************/
  if (argc < 3) Help (argv[0]);
  
  numberPaths = atoi (argv[2]);
  
  for (i=3; i < argc; i++) {
    if (strcmp(argv[i], "-paths") == 0) showPaths = 1;
    else if (strcmp(argv[i], "-tdijkstra") == 0)  measureDijkstra = 1;
    else Help(argv[0]);
  }
  
  /****************** Prints experimental trace information *****************/
  //gethostname (hostName, 200);
  date = time (NULL);
  localDate = localtime (&date);
 // printf ("=========================================================================\n");
  //printf ("CommandLine: ");
 // for (i = 0; i < argc; i++)
 //   printf (" %s", argv[i]);
 // printf ("\nHostname: %s", hostName);
 // printf ("\nDate: %s", asctime(localDate));
 // printf ("=========================================================================\n");
  
  /******************* Reads the graph from file ****************************/
  
  LoadGraph(&graph, argv[1]);
  
  /******************* Dynamic memory allocation *******************************/
  
  /* Memory for heaps is allocated for the worst case */
  heap3Size = graph.numberArcs - graph.numberNodes + 1
              + graph.numberNodes*log(graph.numberNodes)/log(2.0) + 1;
  firstFreeHeap3Node = heap3 = malloc(sizeof(heap3[0])*heap3Size);
  if (heap3 == NULL) {
    perror("Not enough memory for heap3\n");
    exit(1);
  }
  
  cumulatedSeconds = malloc (sizeof(cumulatedSeconds[0])*numberPaths);
  if (cumulatedSeconds == NULL) {
    perror("Not enough memory for time counters.\n");
    exit(1);
  }
  
  stack = malloc (sizeof(stack[0]) * graph.numberNodes);
  if (stack == NULL) {
    perror("Not enough memory for auxiliary stack.\n");
    exit(1);
  }
  traceBackStack = malloc (sizeof(traceBackStack[0]) * numberPaths);
  if (traceBackStack == NULL) {
    perror("Not enough memory for auxiliary stack.\n");
    exit(1);
  }
  
  firstFreeDevPath = devPaths = malloc(sizeof(devPaths[0]) * 4 * numberPaths);
  /* Worst case: 4 insertions for each path */
  if (devPaths == NULL) {
    perror("Not enough memory for devPaths\n");
    exit(1);
  }
  
  searchHeap = malloc(sizeof(searchHeap[0]) * 3 * numberPaths);
  /* Worst case:  1 deletion and 4 insertions for each path */
  if (searchHeap == NULL) {
    perror("Not enough memory for searchHeap\n");
    exit(1);
  }
  
  bestDevPaths = malloc(sizeof(bestDevPaths[0]) * numberPaths);
  if (bestDevPaths== NULL) {
    perror("Not enough memory for bestDevPaths\n");
    exit(1);
  }
  
  /****************** Computes the shortest paths tree **********************/
  if (measureDijkstra == 1)
    ClockReset ();
  
  DijkstraReversed (&graph);
  
  if (measureDijkstra == 1)
    cumulatedSeconds[0] =  ClockTotal();
  else {
    cumulatedSeconds[0] =  0;
    ClockReset ();
  }
  
#ifdef DEBUG
  for (node = graph.node, numberNodes = graph.numberNodes; numberNodes != 0; 
       node++, numberNodes--) { 
    printf ("\nBest Path for InitialNode=%i: ", node->name);
    PrintBestPath (node);
  }
#endif
  
  /******************** Computes the K shortest paths ***********************/
  
  /**************** Builds the heaps of path deviations *********************/

  BuildDeviationHeaps(&graph);
  
  /************ Searches the heaps of path deviations ***********************/
  SearchDeviationPaths (&graph, numberPaths);
  
  /********************* Makes the best paths explicit ***********************/
  /************ Prints the computed paths and time counters ******************/
  if( !(outf = fopen("tmpeppout", "w")) )
	  fprintf(stdout, "File tmpeppout cannot be opened for writing!\n"), fflush(NULL), exit(0);
  
  for(i=0; i<numberComputedPaths && i < numberPaths; i++) {
    int top = -1;
    DevPath * p;
    fprintf (outf, "\nN=%i:\t", i+1);
    for (p = bestDevPaths[i]; p!=NULL; p=p->parent) traceBackStack[++top] = p;
    node = graph.initialNode;
    if (showPaths) fprintf (outf, "%d-", node->name);
    while (top >= 0 || node != graph.finalNode) {
      if (top>=0 && node == traceBackStack[top]->heap3->data.arc->source) 
	node = traceBackStack[top--]->heap3->data.arc->dest;
      else
	node = node->bestPath;
      if (showPaths) fprintf (outf, "%d-", node->name);
    }            
    fprintf (outf, " \t(Cost: %7.2f)", graph.initialNode->bestCost + ((i>0) ? bestDevPaths[i]->cost : 0));
    if (measureDijkstra == 1)
      fprintf (outf, " \t(CumulatedSeconds: %.2f)", (float) cumulatedSeconds[i]);
  }
  
  if (measureDijkstra == 1)
    fprintf (outf, "\nTotalExecutionTime: %.2f\n", (float) cumulatedSeconds[numberComputedPaths-1]);

  fprintf (outf, "\n\n");
  fclose(outf);
  
  return (0);
  
}
