/*========================================================================

  File: EPPSTEIN.h

  Program EPPSTEIN - Version 1.1 - July 1999

  ========================================================================

  This module contains an implementation of Eppstein's Algorithm
  that enumerates (by increasing weight) the N shortest paths in
  weighted graphs. The algorithm is described in: 

         "Finding the K Shortest Paths" by David Eppstein,
         SIAM J. COmput. vol. 28 no. 2 pp. 652-673 (1999)

  This implementation corresponds to  what is called the basic version
  in this paper. After the  shortest paths tree has been computed, the
  K shortest  paths are enumerated  in time  O(m+ n log  n + K  log K)
  where n  is the  number of  nodes and m  the number  of arcs  in the
  graph.

  ========================================================================

    Copyright (C) 1999 Victor Jimenez and Andres Marzal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (file COPYING); if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    You can contact the authors at:

           Victor Jimenez and Andres Marzal
           Dept. de Informatica, Universitat Jaume I, Castellon, Spain
           {vjimenez,amarzal}@inf.uji.es

    Updated information about this software will be available at the
    following www address:

           http://terra.act.uji.es/REA
 
  ========================================================================= */

#ifndef _EPPSTEIN_H_INCLUDED

#define COST_TYPE double     /* You can change this type but you should verify
                             its use in printf's, scanf's and castings */

#define INFINITY_COST 10000000.0 /* Instead of INT_MAX, to avoid overflow  */
                                /* when adding some cost to INFINITY_COST */


/* The following data structures represent a graph in memory providing
   access to incoming and outgoing arcs for each node, and also allow
   to represent the shortest paths from every node to the final node
   and the heap of deviation paths associated to each node.
*/

typedef struct Heap3Data {
  COST_TYPE  cost;
  struct Arc *arc;
} Heap3Data;

typedef struct Heap3Node {
  Heap3Data        data;
  int              desc;     /* Number of descendents, to balance the heaps */
  struct Heap3Node *left,
                   *center,
                   *right;   /* Childs */
} Heap3Node;

typedef struct Arc {
  COST_TYPE   cost;     /* Arc cost         */
  struct Node *source;  /* Source node      */
  struct Node *dest;    /* Destination node */
} Arc;  

typedef struct Arc *PtrArc;

typedef struct Node {
  int         name;          /* An integer in the range 1..graph.numberNodes  */
  int         numberArcsOut; /* Number of arcs that leave the node            */
  PtrArc      *firstArcOut;  /* Pointer to an element in graph.arcOut         */
  int         numberArcsIn;  /* Number of arcs that reach the node            */
  Arc         *firstArcIn;   /* Pointer to an element in graph.arc            */
  Heap3Node   *heap3;        /* Heap of deviation paths                       */
  int         heap3Done;     /* Flag indicating if the heap has been created  */
  COST_TYPE   bestCost;      /* Cost of the best path from this node          */
                             /* to the final node.                            */
  struct Node *bestPath;     /* Next node in the best path                    */
} Node;

typedef struct Node *PtrNode;

typedef struct Graph {
  int    numberNodes;  /* Number of nodes in the graph                        */
  int    numberArcs;   /* Number of arcs in the graph                         */
  Node   *initialNode; /* The node from which all paths depart                */
  Node   *finalNode;   /* The node to which the K shortest paths are required */
  Node   *node;        /* Nodes, with node named i in position i-1            */
  Arc    *arc;         /* Arcs sorted by destination node                     */
  PtrArc *arcOut;      /* Pointers to arcs, sorted by source node             */
}  Graph;

#define _EPPSTEIN_H_INCLUDED
#endif


