/*
 Author: Qingwu Yang
 Department of Computer Science
 Texas A&M University, Texas USA
 Mar, 2006
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>
#include <float.h>
#include <ctype.h>
#include <math.h>
#include <time.h>

#define BOOLEAN char
#define TRUE 1
#define FALSE 0

#define LONGEST_SEQ_LEN 40000
#define LENGTH_LONGEST_SHORTEST_PATH 100 // The length (number of edges) of the longest possible shortest path

char **ARGV;
int    ARGC;
//>>>>> modifiable parameters >>>>>

int if_directed_graph = 0;

int NO_OF_TOP_PATHS_OUTPUT = 10; // Number of pathways with top scores that are required to be output

char seqoutfname1[50]; //query file
char network_file[100]; // file containing the network 
char corrfilename[50];   //corr file name
char topoutfname[50];   //result file name


//cutoffs for recording bl2seq results
double E_VALUE_CUTOFF  = 1.0; //1e-2
int    E_LENGTH_CUTOFF =  0;

//cutoffs for determining similarity
double CORR_E_VALUE_CUTOFF= 1e200; //1e-5;
int    CORR_E_LENGTH_CUTOFF  = 0;

//cutoffs for determining similarity according to EC
double CORR_EC_CUTOFF=-1.7;

//rxnnetwork
int    RXNNETWORK_SIMILARITY_USING_COMPSEARCH_INFORMATIONCONTENT = 1;
int    RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS = 1;

//<<<<< modifiable parameters <<<<<

/* maximum number of mismatches and gaps allowed */
int MAX_NO_OF_MISMATCHES_AND_GAPS=1;

int NETWORK_TYPE =1; /* 0 -- protein interaction network;
                        1 -- reaction network; */

int GENERAL=1; /*GENERAL=0, the first and last nodes of the input pathway must have correspondences;
                 GENERAL=1, gaps allowed at the beginning and end of the input pathway.             */
int LOOPEDGE=0; /*LOOPEDGE=1, there is a loop edge for each node;
                  LOOPEDGE=0, no loop edge for each node*/

int if_de_novo_graph1=1; //

int if_mask_input_nodes=1; // if the graph1 contains the input nodes, they will be masked.
int if_blast2seq_pathwayquery=1;
int if_shrt_whole_graph1=1; //1 shortest paths on the whole graph1; 0 -- only correspondence pairs

/* penalty for mismatches and gaps*/
double INDEL_PENALTY = 1.0;

typedef struct node {
	int uid;
	int id;
	char * name;
	char * clss;
	char * swp;
	char * pir;
	char *  gi;
	char * organism;
	char   thisorganism; // 1 right; 0, other organism.
	char * shn;
	char   shnindex; // 1 there exits a shn; 0, otherwise.
	char * descr;
	char * pid; //HP0587  protein id in an organism
	char * pname;  //parsed name
} node;

typedef struct edge {
	int uid;
	int  id;
	int node1; //id of node
	int node2; //id of node
} edge;


//>>>>>>>>> path searching >>>>>>>>>

//hold the shortest path recovered
int spath[LENGTH_LONGEST_SHORTEST_PATH];
int spath_count;

//hold the final path recovered
int finalpath0[LENGTH_LONGEST_SHORTEST_PATH];
int finalpath1[LENGTH_LONGEST_SHORTEST_PATH];
int finalpath2[LENGTH_LONGEST_SHORTEST_PATH];
int finalpath_count0;
int finalpath_count1;
int finalpath_count2;
//int finalpathscore;

//allocate space for graph
void allocate_space(int ** g, int size);

//free space of graph
void free_space(int ** g, int size);

//floyd's algorithm
void floyd(int ** g, int ** dist, int ** path, int size);

//test negative cycle
char test_negative_cycle(int ** dist, int size);

//recover the shortest path
void output_all_pairs_shortest_path(int ** path, int start, int end);

//print the distance matrix
void print_dist_matrix(int ** dist, int size, char * fname);

//print the path matrix
void print_path_matrix(int ** path, int size, char * fname);

//read in the path matrix
void read_path_matrix(int ** dist, int size, char * fname);

//read in the distance matrix
void read_dist_matrix(int ** path, int size, char * fname);

//<<<<<<<<< path searching <<<<<<<<<

//>>>>>>>> correspondences >>>>>>
int NUM_CORRESPONDENCES;
typedef struct pair {
	int v1; //vertex in graph1
	int v2; //vertex in graph2
	int id1; //PIN node ID for v1
	int id2; //PIN node ID for v2
	double  sim;      //similarity or other measurement
} pair;

//allocate space for corr
void alloc_space(pair * co, int num);

//read in correspondence information
void correspondence_reader(pair ** co, char * fname, int * num, int n, char **names, char **qnames, int q_size);

//generate the correspondence graph. vertices--correspondence pairs; edge -- length of shortest paths
void generate_corr_graph(int ** dist1, int ** dist2, pair * co, int ** g, int n1, int n2, int n3);

//generate a correspondence file from ec numbers for rxnnetwork
void create_corr_file_ec_rxnnetwork(char ** inenzyme, int innum, node * ns, int nsnum, char * corrfname);
//<<<<<<<< correspondences <<<<<<

//******pathway query related*********>
//read in the pathway to be queried. return 1 if success; 0, otherwise
BOOLEAN pathway_query_read_path(char * fname, char ** inpname, int * inplen);

//read in the pathway to be queried. return 1 if success; 0, otherwise
BOOLEAN pathway_query_read_path_FASTA(char * fname, char ** inpname, int * inplen);

//read in the metabolic enzyme pathway to be queried. return 1 if success; 0, otherwise
BOOLEAN pathway_query_read_path_EC(char * fname, char ** inpname, int * inplen);

//generate correspondences for nodes in the pathway queried and calculate the log-odds values
void pathway_query_create_corr2(int * inpid, int inplen, int flag, pair * co, int cono,
	                           int **inpcorr, double **inpcorrwt, int * inpcorrsize, double scutoff);

//calculate orders for the dag with correspondences as vertices
void pathway_query_calculate_orders(int inplen, int *inpcosize, int * cdorder, int * cdlorder);

//allocate memory for the dbldag
void pathway_query_allocate_mem2(double ** dag, int inplen, int *inpcosize, int cdorder, int * cdlorder);

//calculate weights for corrdag, mismatch and indel penalty is considered. Upper bound for sum of mismatches and indels is applied.
void pathway_query_calculate_weight4(double ** dag, int inplen, int ** inpco, double ** inpcowt, int *inpcosize, int cdorder,
									int * cdlorder, int ** dist, int nno);

//Recover a shortest path
void pathway_query_recover_path(int inplen, int ** inpco, int *inpcosize,
								 int ** pth, int nno, node *no, char **inpname, int *inpid);

//find k pathways with top scores.
//This function always checks whether the score of a new path is within current top k scores.
void pathway_query_top_paths3(double ** dag, int inplen, int ** inpco, double ** inpcowt, int *inpcosize, int cdorder,
							 int * cdlorder, int ** pth, int nno, node *no, char **inpname, int *inpid,
							 pair * co, int cono, int flag, char *outfname);

//******pathway query related*********<
typedef struct {
	int item;   //vertex id
	int weight; //weight
} vert;

//DOT related
typedef struct pathvert {
	double score;
	int len;
	int * path;
	struct pathvert * next;
	int * mask;// [2*NO_OF_TOP_PATHS_OUTPUT]; edges masked
	int maskno;
	char ifcut; //0 keep the initial edges between start and end  1 remove otherwise
} pathvert;

//show commandline parameters
void infobox();

//read graphs in linked list representation and save to adjacent matrix
void graph_reader_1(char *f, int *** g, int gsize, char ** names);

//given a graph, return number of edges
int count_graph_edge(int ** g, int n);

//read in node names
void nname_reader(char *fname, int * size, char ***nnames);

//output network in new format
void output_network_for_test(char *fname, int ** network, int network_size, char **networknodenames);
//output correspondence in new format
void output_correspondence_for_test(char *fname, pair * corr, int n, char **names);

int main(int argc, char ** argv){
	//FILE * tim; //, seqf;
	node  * nodes2;//* nodes1,
	//char seqoutfname1[50];
	//char topoutfname[50]; //shrtdistname1[50], shrtpathname1[50],
	int i;// j,x;
	double simcutoff;

	/* variables related to pathway query */
	char pathqueryoutfilename[50]; //file where results of path_query to be stored.
	char *inpathname[LENGTH_LONGEST_SHORTEST_PATH]; //array to hold names of nodes in the input pathway
	int   inpathlen;              // length(number of nodes) of the input pathway
	int * inpathid;               //array to hold indices of nodes in the input pathway
	int **inpathcorr;             //array to hold correspondences of the nodes in the input pathway
	int * inpathcorrsize;         //array to hold the size of correspondences of the nodes in the input pathway
	double **inpathcorrwt;         //array to hold the log-odds
	double ** dblcodag;
	int   codagorder;             //total number of nodes in the dag
	int  *codaglevelorder;        //number of edges for nodes on different levels

	//time_t  timer;
	//clock_t ticks0, ticks1, ticks2;

	//***** path searching ******>
	int SIZE_GRAPH2; // number of vertices in graph 2

	int ** graph2;   // graph 2 (represented in adjacent matrix)

	//char corrfilename[50];
	pair * corr;  //hold the correspondence information

	//network graph
	//char   network_file[100]; // file containing the network 
	int ** network;       // queried graph (represented in adjacent matrix)
	int    network_size;  // number of nodes in network
	char   network_shrt_dist_fname[100], network_shrt_path_fname[100];
	int ** network_shrt_dist;
	int ** network_shrt_path;
	char **networknodenames;
	char   networknodenames_file[100];

	//if( !(tim=fopen("log", "w")) )
	//	fprintf(stdout, "can't open the time file for writing!\n"), fflush(NULL), exit(0);

	ARGV = argv;
	ARGC = argc;

	strcpy(topoutfname, "result");
	strcpy(seqoutfname1, "query");
	strcpy(corrfilename,   "corr");

	strcpy(network_file,  "input");
	strcpy(networknodenames_file, "networknames");

	//command line input processing
	if( argc < 2 ){
		infobox();
		exit(0);
	} else {
		for( i=1; i< argc; i++ ){
			if( strstr(argv[i], "-g") ) {
				if( sscanf(argv[i], "-g=%d", &MAX_NO_OF_MISMATCHES_AND_GAPS ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			} else if( strstr(argv[i], "-s") ) {
				if( sscanf(argv[i], "-s=%d", &GENERAL ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			} else if( strstr(argv[i], "-l") ) {
				if( sscanf(argv[i], "-l=%d", &if_blast2seq_pathwayquery ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			} else if( strstr(argv[i], "-v") ) {
				if( sscanf(argv[i], "-v=%d", &if_de_novo_graph1 ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			}  else if( strstr(argv[i], "-n") ) {
				if( sscanf(argv[i], "-n=%d", &NO_OF_TOP_PATHS_OUTPUT ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			} else if( strstr(argv[i], "-b") ) {
				if( sscanf(argv[i], "-b=%le", &E_VALUE_CUTOFF ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			} else if( strstr(argv[i], "-r") ) {
				if( sscanf(argv[i], "-r=%le", &CORR_E_VALUE_CUTOFF ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			} else if( strstr(argv[i], "-e") ) {
				if( sscanf(argv[i], "-e=%le", &CORR_EC_CUTOFF ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			} else if( strstr(argv[i], "-i") ) {
				if( sscanf(argv[i], "-i=%s", network_file ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			}  else if( strstr(argv[i], "-j") ) {
				if( sscanf(argv[i], "-j=%s", networknodenames_file ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			} else if( strstr(argv[i], "-o") ) {
				if( sscanf(argv[i], "-o=%s", pathqueryoutfilename ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
				strcpy(topoutfname, pathqueryoutfilename);
			} else if( strstr(argv[i], "-q") ) {
				if( sscanf(argv[i], "-q=%s", seqoutfname1 ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			}  else if( strstr(argv[i], "-c") ) {
				if( sscanf(argv[i], "-c=%s", corrfilename ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}
			} else if( strstr(argv[i], "-p") ) {
				if( sscanf(argv[i], "-p=%le", &INDEL_PENALTY ) != 1 ){
					fprintf(stdout, "\nwrong command line input:\n\t%s\n", argv[i]);
					infobox();
					exit(0);
				}

				INDEL_PENALTY *= (-1);
			} else {
				fprintf(stdout, "\n!wrong command line input:\n\t%s!\n", argv[i]);
				infobox();
				exit(0);
			}
		}
	}

	strcpy(network_shrt_dist_fname, "input_shrt_dist");
	strcpy(network_shrt_path_fname, "input_shrt_path");

	//read in node names of the network
	nname_reader(network_file, &network_size, &networknodenames);

	//printf("\n\nnetwork_size = %d\n\n", network_size), fflush(NULL);

	//read network
	graph_reader_1(network_file, &network, network_size, networknodenames);

	//output network in new format
	//output_network_for_test("network_test", network, network_size, networknodenames);

	if( !( nodes2 = malloc(network_size * sizeof(node))) )
		fprintf(stderr, "\nout of memory!\n\n"), fflush(NULL), exit(0);

	for(i=0; i<network_size; i++)
		nodes2[i].shn = nodes2[i].name = networknodenames[i];

	graph2 = network;
	SIZE_GRAPH2 = network_size;

	if( if_de_novo_graph1==1 ){ //apply shortest paths algorithm on graph1
		//allocate space for matrices related to the shortest paths of graphs
		if( !(network_shrt_dist=malloc(network_size*sizeof(int *))) )
			fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);
		allocate_space(network_shrt_dist, network_size);
		if( !(network_shrt_path=malloc(network_size*sizeof(int *))) )
			fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);
		allocate_space(network_shrt_path, network_size);

		//floyd's algorithm to calculate the all-pairs shortest paths		
		floyd(graph2, network_shrt_dist, network_shrt_path, network_size);

		//print_dist_matrix(network_shrt_dist, network_size, network_shrt_dist_fname);
		//print_path_matrix(network_shrt_path, network_size, network_shrt_path_fname);		
	} else {  //don't apply shortest paths algorithm on graph1. load directly
		//create graph2
		//SIZE_GRAPH2=network_size;

		//allocate space for matrices related to the shortest paths of graphs
		if( !(network_shrt_dist=malloc(network_size*sizeof(int *))) )
			fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);
		allocate_space(network_shrt_dist, network_size);
		if( !(network_shrt_path=malloc(network_size*sizeof(int *))) )
			fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);
		allocate_space(network_shrt_path, network_size);

		read_dist_matrix(network_shrt_dist, network_size, network_shrt_dist_fname);
		read_path_matrix(network_shrt_path, network_size, network_shrt_path_fname);
	}

	//read in the pathway to be queried. return 1 if success; 0, otherwise
	if( !(pathway_query_read_path_FASTA(seqoutfname1, inpathname, &inpathlen)) ){
		fprintf(stdout,"cannot read in the pathway queried correctly!\n");
		fflush(stdout);
	}

	//read in the correspondence information
	correspondence_reader(&corr, corrfilename,  &NUM_CORRESPONDENCES, network_size, networknodenames, inpathname, inpathlen);

	//output correspondence in new format
	//output_correspondence_for_test("corr_test", corr, NUM_CORRESPONDENCES, networknodenames);

	//convert the names of nodes into indices
	if( !(inpathid = malloc(inpathlen*sizeof(int))) ) {
		fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
	}
	for(i=0;i<inpathlen;i++)
		inpathid[i]=i;

	//generate correspondences for nodes in the pathway queried
	if( !(inpathcorrsize = malloc(inpathlen*sizeof(int))) )
		fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
	if( !(inpathcorr     = malloc(inpathlen*sizeof(int *))) )
		fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
	if( !(inpathcorrwt   = malloc(inpathlen*sizeof(double *))) )
		fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
	if( !(codaglevelorder = calloc(inpathlen, sizeof(int))) )
		fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);

	INDEL_PENALTY  = -INDEL_PENALTY;

	simcutoff=CORR_E_VALUE_CUTOFF;
	pathway_query_create_corr2(inpathid, inpathlen, 1, corr, NUM_CORRESPONDENCES,
								inpathcorr, inpathcorrwt, inpathcorrsize, simcutoff);

	//create a dag with correspondences as vertices + sink +source
	pathway_query_calculate_orders(inpathlen, inpathcorrsize, &codagorder, codaglevelorder);

	//allocate memory for the dag
	if( !(dblcodag = malloc(codagorder*sizeof(double *))) )
		fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
	pathway_query_allocate_mem2(dblcodag, inpathlen, inpathcorrsize, codagorder, codaglevelorder);

	//calculate edge weights
	pathway_query_calculate_weight4(dblcodag, inpathlen, inpathcorr, inpathcorrwt, inpathcorrsize,
										codagorder, codaglevelorder, network_shrt_dist, network_size);
	
	//calculate the shortest path
	pathway_query_top_paths3(dblcodag, inpathlen, inpathcorr, inpathcorrwt, inpathcorrsize, codagorder,codaglevelorder, 
			network_shrt_path, network_size, nodes2, inpathname, inpathid, corr, NUM_CORRESPONDENCES, 1, topoutfname);					
	
	//timer=time(NULL);
	//fprintf(tim, "Time Done: %s\n", asctime(localtime(&timer)));

	//ticks2=clock();
	//fprintf(tim, "TOTAL CPU TIME:  %8.2f  SECs / %8.2f  MINs / %8.2f  Hrs \n",
	//		(double) (ticks2-ticks0)/CLOCKS_PER_SEC, (double) (ticks2-ticks0)/CLOCKS_PER_SEC/60, (double) (ticks2-ticks0)/CLOCKS_PER_SEC/3600);
	//fclose(tim);

	remove("tmpeppout");
	remove("tmpgraph");

	return 0;
} //main()

//output correspondence in new format
void output_correspondence_for_test(char *fname, pair * corr, int n, char **names)
{
	FILE * outf;
	int i;

	outf = fopen(fname, "w");

	for(i=0; i<n; ++i){
		fprintf(outf, "\n%5d %5d %le", corr[i].v1 +1, corr[i].v2 +1, corr[i].sim);	
	}

	fclose(outf);

	return;
}

//output network in new format
void output_network_for_test(char *fname, int ** network, int network_size, char **networknodenames)
{
	FILE * outf;
	int i, j;

	outf = fopen(fname, "w");

	for(i=0; i<network_size; ++i){
		fprintf(outf, "\n%1d, ", i+1);

		for(j=0; j<network_size; ++j){
			if( network[i][j] == 1 )
				fprintf(outf, "%1d, ", j+1);
		}
	}

	fclose(outf);

	return;
}


void string2lowcase(char * temp){
	int i, len;

	len=(int)strlen(temp);
	for(i=0;i<len;i++)
		temp[i]=tolower(temp[i]);
}


//check if one line is blank
int is_blank(char *s){
	while( *s != '\n' && *s != '\0' && *s != EOF ){
		if( isalnum( *s ) )
			return 0;
		s++;
	}

	return 1;
}

//******pathway query related*********>
//read in the pathway to be queried. return 1 if success; 0, otherwise
BOOLEAN pathway_query_read_path(char * fname, char ** inpname, int * inplen){
	FILE * inf;
	int i, x;
	char buff[4096], buffer[4096];
	char *token;

	if( !(inf=fopen(fname, "r")) ){
		fprintf(stdout, "%1s cannot be opened for reading!\n", fname);
		fflush(stdout);
		return 0;
	}

	while( fgets(buff, 4096, inf) && (strlen(buff) < 3)){}  //skip the blank lines and the organism name line
	while( fgets(buff, 4096, inf) ){
		if(strlen(buff) >=3){
			i=0;
			strcpy(buffer, buff);

			//determine the number of nodes in the pathway queried
			if( (token=strtok(buffer, ", ")) )	i++;
			while( (token=strtok(NULL, ", ")) )	i++;
			*inplen=i;

			//read in the name of nodes in the pathway queried
			strcpy(buffer, buff);

			i=0;
			if( (token=strtok(buffer, ", ")) ) {
				if( !(inpname[i]=malloc((strlen(token)+1)*sizeof(char))) ){
					fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
				}
				strcpy(inpname[i], token);
				i++;
			}

			while( (token=strtok(NULL, ", ")) ) {
				if( !(inpname[i]=malloc((strlen(token)+1)*sizeof(char))) ){
					fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
				}
				strcpy(inpname[i], token);
				i++;
			}

			for(x=0;x<(int)strlen(inpname[i-1]);x++){
				if( !(isalnum((unsigned char)inpname[i-1][x])) ){
					inpname[i-1][x]='\0';
				}
			}
		}
	}

	fclose(inf);
	return 1;
}

//read in the pathway to be queried. return 1 if success; 0, otherwise
BOOLEAN pathway_query_read_path_FASTA(char * fname, char ** inpname, int * inplen){
	FILE * inf;
	int i;
	char buff[4096];
	char * ptr, *ptr1;

	if( !(inf=fopen(fname, "r")) ){
		fprintf(stdout, "%1s cannot be opened for reading!\n", fname);
		fflush(stdout);
		return 0;
	}

	i=0;
	while( fgets(buff, 4096, inf) ){
		if( ! is_blank(buff) ){
			ptr=buff;
			while( *ptr == ' ' ) ptr++;
			ptr1=ptr;
			while( *ptr1 != '\0' ){
				if( *ptr1 == '\n' ){
					*ptr1 = '\0';
					break;
				}
				ptr1++;
			}
			if( !(inpname[i]=malloc((strlen(ptr)+1)*sizeof(char))) ){
					fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
			}
			strcpy(inpname[i], ptr);
			i++;
		}
	}

	*inplen=i;

	fclose(inf);

	if(i==0){
		fprintf(stdout, "empty input pathway!\n"), fflush(NULL), exit(0);
	} else
		return 1;
}

//read in the metabolic enzyme pathway to be queried. return 1 if success; 0, otherwise
BOOLEAN pathway_query_read_path_EC(char * fname, char ** inpname, int * inplen){
	FILE * inf;
	int i;
	char buff[4096];
	char * ptr, *ptr1;

	if( !(inf=fopen(fname, "r")) ){
		fprintf(stdout, "%1s cannot be opened for reading!\n", fname);
		fflush(stdout);
		return 0;
	}

	i=0;
	while( fgets(buff, 4096, inf) ){
		if( (ptr=strstr(buff, ">")) ){
			ptr++;
			ptr1=ptr;
			while( *ptr1 != '\0' ){
				if( *ptr1 == '\n' ){
					*ptr1 = '\0';
					break;
				}
				ptr1++;
			}
			if( !(inpname[i]=malloc((strlen(ptr)+1)*sizeof(char))) ){
					fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
			}
			strcpy(inpname[i], ptr);
			i++;
		}
	}

	*inplen=i;

	fclose(inf);

	if(i==0){
		fprintf(stdout, "empty input pathway!\n"), fflush(NULL);
		return 0;
	} else
		return 1;
}

//generate correspondences for nodes in the pathway queried and calculate the log-odds values
void pathway_query_create_corr2(int * inpid, int inplen, int flag, pair * co, int cono,
	                           int **inpcorr, double **inpcorrwt, int * inpcorrsize, double scutoff){
	int i,j, cnt;

	if(flag == 1){
		for(i=0;i<inplen;i++){
			cnt=0;
			for(j=0;j<cono;j++){
				if( (co[j].v1 == inpid[i]) && (co[j].sim <= scutoff))
					cnt ++;
			}
			inpcorrsize[i]=cnt;
		}
		for(i=0;i<inplen;i++){
			if( !(inpcorr[i]=malloc( inpcorrsize[i] * sizeof(int) )) ){
				fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
			}
			if( !(inpcorrwt[i]=malloc( inpcorrsize[i] * sizeof(double) )) ){
				fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
			}
		}
		for(i=0;i<inplen;i++){
			cnt=0;
			for(j=0;j<cono;j++){
				if( (co[j].v1 == inpid[i]) && (co[j].sim <= scutoff) ){
					inpcorr[i][cnt]  =co[j].v2;

					if( if_blast2seq_pathwayquery == 1 ){
						if(co[j].sim < 0){
							fprintf(stdout, "Negative co[i].sim\n"), fflush(stdout), exit(0);
						} else if(co[j].sim <= 1e-299){
							inpcorrwt[i][cnt]=log(1e-299);			
						} else {
							inpcorrwt[i][cnt]=log(co[j].sim);
						}
					} else {
						inpcorrwt[i][cnt]= co[j].sim;
					}

					cnt++;

					if(cnt==inpcorrsize[i]){
						break;
					}
				}
			}
		}
	} else {
		fprintf(stdout, "orgflag = %1d  (neither 1 nor 2)!\n", flag), fflush(stdout), exit(0);
	}
}

//calculate orders for the dag with correspondences as vertices
void pathway_query_calculate_orders(int inplen, int *inpcosize, int * cdorder, int * cdlorder){
	int i;

	*cdorder=0;
    for(i=0;i<inplen;i++)
		*cdorder += inpcosize[i];

    cdlorder[0]=*cdorder-inpcosize[0];
    for(i=1;i<inplen;i++)
		cdlorder[i] = cdlorder[i-1]- inpcosize[i];

	return;
}

//allocate memory for the dbldag
void pathway_query_allocate_mem2(double ** dag, int inplen, int *inpcosize, int cdorder, int * cdlorder){
	int i,j, k;

	k=0;
	for( i=0; i<=inplen-2; i++ ){
		for(j=0;j<inpcosize[i];j++){
			if( !(dag[k++]=calloc(cdlorder[i], sizeof(double))) )
				fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
		}
	}

	return;
}

//calculate weights for corrdag, mismatch and indel penalty is considered. Upper bound for sum of mismatches and indels is applied.
void pathway_query_calculate_weight4(double ** dag, int inplen, int ** inpco, double ** inpcowt, int *inpcosize, int cdorder,
									int * cdlorder, int ** dist, int nno){
	int i,j, x, y, m, n;
	int d1, d2, gap;

	m=0;
	for( i=0; i<=inplen-2; i++ ){
		for(j=0;j<inpcosize[i];j++){
			n=0;
			for(x=i+1;x<=inplen-1;x++){
				d1=x-i;
				for(y=0;y<inpcosize[x];y++){
					if( n>=cdlorder[i] )
						fprintf(stdout, "out of boundary!\n"), fflush(stdout), exit(0);

					d2=dist[inpco[i][j]][inpco[x][y]];
					if( (d1<=(MAX_NO_OF_MISMATCHES_AND_GAPS+1)) && (d2<=(MAX_NO_OF_MISMATCHES_AND_GAPS+1)) ){
						gap=(d2>d1)?(d2-1):(d1-1);
						dag[m][n]=inpcowt[x][y]+INDEL_PENALTY*gap;
					} else
						dag[m][n]=FLT_MAX;

					n++;
				}
			}
			if( n != cdlorder[i] )
				fprintf(stdout, "Less members!\n"), fflush(stdout), exit(0);

			m++;
		}
	}

	return;
}


//Recover a shortest path  //10-15-04 allows gaps at the beginning and end of the pathways
void pathway_query_recover_path(int inplen, int ** inpco, int *inpcosize,
								 int ** pth, int nno, node *no, char **inpname, int *inpid){
	int i,j, x, y, m, len, cond;
	int level[LENGTH_LONGEST_SHORTEST_PATH][2];
	int mark1[LENGTH_LONGEST_SHORTEST_PATH];

	for(i=0;i<spath_count-2;i++){
		m=cond=0;
		for(x=0;x<=inplen-1;x++){
			for(y=0;y<inpcosize[x];y++){
				if(m==spath[i+1]-1){
					level[i][0]=x;
					level[i][1]=y;
					cond=1;
					break;
				} else
					m++;
			}

			if(cond==1)
				break;
		}

	}

	len=finalpath_count0=spath_count-2;
	for(i=0;i<finalpath_count0;i++)
		finalpath0[i]=spath[i+1];

	//create path 1 to be returned
	finalpath_count1=0;
	mark1[0]=0;
	for(i=1;i<finalpath_count0;i++){
		//codes after 10-20-04
		if( inpco[level[i-1][0]][level[i-1][1]] == inpco[level[i][0]][level[i][1]] ){
			if(finalpath_count1 == 0){
				finalpath1[0]=finalpath1[1]=inpco[level[i-1][0]][level[i-1][1]];
				finalpath_count1 = 2;
				mark1[i]=finalpath_count1-1;
			}else {
				if( (finalpath_count1+spath_count-1) > LENGTH_LONGEST_SHORTEST_PATH){
					fprintf(stdout, "LENGTH_LONGEST_SHORTEST_PATH is set too small!\n");
					exit(0);
				}
				finalpath1[finalpath_count1]=inpco[level[i-1][0]][level[i-1][1]];
				finalpath_count1++;
				mark1[i]=finalpath_count1-1;
			}
		} else {
			spath_count=0;
			output_all_pairs_shortest_path(pth, inpco[level[i-1][0]][level[i-1][1]], inpco[level[i][0]][level[i][1]]);
			if(finalpath_count1 == 0){
				for(j=0; j<spath_count; j++)
					finalpath1[finalpath_count1+j]=spath[j];

				finalpath_count1 = spath_count;
				mark1[i]=finalpath_count1-1;
			}else {
				if( (finalpath_count1+spath_count-1) > LENGTH_LONGEST_SHORTEST_PATH){
					fprintf(stdout, "LENGTH_LONGEST_SHORTEST_PATH is set too small!\n");
					exit(0);
				}
				for(j=1; j<spath_count; j++)
					finalpath1[finalpath_count1+j-1]=spath[j];

				finalpath_count1 += (spath_count-1);
				mark1[i]=finalpath_count1-1;
			}
		}	
	}

	finalpath_count0=finalpath_count1;
	for(i=0;i<finalpath_count0;i++)
		finalpath0[i]=finalpath1[i];

	finalpath_count1=finalpath_count2=x=0;
	//>new
	for(i=0;i<level[0][0];i++){
		finalpath1[finalpath_count1++]=i;
		finalpath2[finalpath_count2++]=-1;
	}
	finalpath1[finalpath_count1]=level[0][0];
	finalpath2[finalpath_count2]=finalpath0[mark1[0]];
	//<new

	for(i=1;i<len;i++){
		if( (level[i][0]-level[i-1][0]) > (mark1[i]-mark1[i-1]) ){
			for(j=1;j<(mark1[i]-mark1[i-1]);j++){
				finalpath1[++finalpath_count1]=level[i-1][0]+j;
				finalpath2[++finalpath_count2]=finalpath0[mark1[i-1]+j];
			}
			for(j=(mark1[i]-mark1[i-1]);j<(level[i][0]-level[i-1][0]);j++){
				finalpath1[++finalpath_count1]=level[i-1][0]+j;
				finalpath2[++finalpath_count2]=-1;
			}
			finalpath1[++finalpath_count1]=level[i][0];
			finalpath2[++finalpath_count2]=finalpath0[mark1[i]];
		} else if ( (level[i][0]-level[i-1][0]) == (mark1[i]-mark1[i-1]) ){
			for(j=1;j<=(mark1[i]-mark1[i-1]);j++){
				finalpath1[++finalpath_count1]=level[i-1][0]+j;
				finalpath2[++finalpath_count2]=finalpath0[mark1[i-1]+j];
			}
		} else {
			for(j=1;j<(level[i][0]-level[i-1][0]);j++){
				finalpath1[++finalpath_count1]=level[i-1][0]+j;
				finalpath2[++finalpath_count2]=finalpath0[mark1[i-1]+j];
			}
			for(j=(level[i][0]-level[i-1][0]);j<(mark1[i]-mark1[i-1]);j++){
				finalpath1[++finalpath_count1]=-1;
				finalpath2[++finalpath_count2]=finalpath0[mark1[i-1]+j];
			}
			finalpath1[++finalpath_count1]=level[i][0];
			finalpath2[++finalpath_count2]=finalpath0[mark1[i]];
		}
	}
	++finalpath_count1;
	++finalpath_count2;

	//>new
	for(i=level[len-1][0]+1;i<inplen;i++){
		finalpath1[finalpath_count1++]=i;
		finalpath2[finalpath_count2++]=-1;
	}
	//<new

	return;
}
//******pathway query related*********<

//>********* path searching ********>

//generate a correspondence file from ec numbers
void create_corr_file_ec_rxnnetwork(char ** inenzyme, int innum, node * ns, int nsnum, char * corrfname){
	FILE *outf, *tmp, *icallf, *iccof;
	int ** inenz;
	int *  inenzid;
	int ** nsec;
	int *  occur, toccur;
	int i,j, k, ch, oc, cnt=0, inenzno;
	double s, score;
	char buff1[4096];
	char *ptr;

	inenzno = 0;
	for(i=0; i<innum; i++){ //assume each enzyme is in format x.x.x.x strictly
		ptr = inenzyme[i];
		j=0;
		while( *ptr != '\0' ){
			if( *ptr == '.' )
				j++;
			ptr++;
		}
		if( j == 3 )
			inenzno++;
	}
	if( !(inenz=malloc(inenzno*sizeof(int *))) ){
		fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);
	}
	if( !(inenzid=malloc(inenzno*sizeof(int))) ){
		fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);
	}

	for(i=0, k=0; i<innum; i++){ //assume each enzyme is in format x.x.x.x strictly
		ptr = inenzyme[i];
		j=0;
		while( *ptr != '\0' ){
			if( *ptr == '.' )
				j++;
			ptr++;
		}
		if( j == 3 ){
			if( !(inenz[k]=malloc(4*sizeof(int))) )
				fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);

			if( sscanf(inenzyme[i], "%d.%d.%d.%d", &inenz[k][0], &inenz[k][1], &inenz[k][2], &inenz[k][3]) == 4){
			}	else {
				fprintf(stdout, "data error in the %d line of the input pathway\n", i+1), fflush(NULL), exit(0);
			}

			inenzid[k] = i + 1;

			k++;
		}
	}

	if( !(nsec=malloc(nsnum*sizeof(int *))) ){
		fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);
	}
	for(i=0;i<nsnum;i++){
		if( !(nsec[i]=malloc(4*sizeof(int))) )
			fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);

		if( sscanf(ns[i].name, "%d.%d.%d.%d", &nsec[i][0], &nsec[i][1], &nsec[i][2], &nsec[i][3]) == 4){
		} else {
			fprintf(stdout, "data error in the %d line of the input pathway\n", i+1), fflush(NULL), exit(0);
		}
	}

	if( !(occur=malloc(nsnum*sizeof(int))) ){
		fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);
	}
	toccur=0;
	for(i=0;i<nsnum;i++){
		occur[i]=1;
		toccur += occur[i];
	}

	tmp=tmpfile();

	if( !(outf=fopen(corrfname, "w")) )
		fprintf(stdout, "%s file cannot be opened for writing!\n", corrfname), exit(0);

	if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 ){
		if( !(icallf = fopen("allicscores", "w")) )
			fprintf(stdout,  "File allicscores cannot be opened for writing!\n");
		if( !(iccof = fopen("allcoscores", "w")) )
			fprintf(stdout,  "File allcoscores cannot be opened for writing!\n");
	}

	cnt=0;
	for(i=0;i<inenzno;i++){
		for(j=0;j<nsnum;j++){
			if( inenz[i][0] != nsec[j][0] ){
				score=log((double)nsnum)/log(2.0);
				s=exp(score);

				if(score <= CORR_EC_CUTOFF) {
					//fprintf(tmp, "%4d %4d %4d %e\n", inenzid[i], ns[j].uid, ns[j].uid, s);
					fprintf(tmp, "%4d %4d %e\n", inenzid[i], ns[j].uid, s);

					if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 ){
						fprintf(iccof, "%4d %4d %12s %12s %e %e\n", inenzid[i], ns[j].uid, inenzyme[inenzid[i]-1], ns[j].name , s, score);
					}

					//fprintf(tmp, "%4d %4d %4d %e %e\n", i+1, j+1, j+1, s, score);
					cnt++;
				}

				if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 ){
					fprintf(icallf, "%4d %4d %12s %12s %e %e\n", inenzid[i], ns[j].uid, inenzyme[inenzid[i]-1], ns[j].name , s, score);
				}

				//score=exp(log(1/nsnum)/log(2));
				//fprintf(tmp, "%4d %4d %4d %e\n", i+1, j+1, j+1, score);
			} else {
				if( inenz[i][1] != nsec[j][1] ){
					ch=oc=0;
					for(k=0;k<nsnum;k++){
						if(inenz[i][0] == nsec[k][0]){
							ch++;
							oc += occur[k];
						}
					}
					score=log(((double)ch*(double)oc)/((double)toccur))/log(2.0);
					s=exp(score);
					if(score <= CORR_EC_CUTOFF) {
						//fprintf(tmp, "%4d %4d %4d %e\n", inenzid[i], ns[j].uid, ns[j].uid, s);
						fprintf(tmp, "%4d %4d %e\n", inenzid[i], ns[j].uid, s);

						if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 )
							fprintf(iccof, "%4d %4d %12s %12s %e %e\n", inenzid[i], ns[j].uid, inenzyme[inenzid[i]-1], ns[j].name , s, score);

						//fprintf(tmp, "%4d %4d %4d %e %e\n", i+1, j+1, j+1, s, score);
						cnt++;
					}

					if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 ){
						fprintf(icallf, "%4d %4d %12s %12s %e %e\n", inenzid[i], ns[j].uid, inenzyme[inenzid[i]-1], ns[j].name , s, score);
					}
				} else {
					if( inenz[i][2] != nsec[j][2] ){
						ch=oc=0;
						for(k=0;k<nsnum;k++){
							if( (inenz[i][0] == nsec[k][0]) && (inenz[i][1] == nsec[k][1]) ){
								ch++;
								oc += occur[k];
							}
						}
						score=log(((double)ch*(double)oc)/((double)toccur))/log(2.0);
						s=exp(score);
						if(score <= CORR_EC_CUTOFF) {
							//fprintf(tmp, "%4d %4d %4d %e\n", inenzid[i], ns[j].uid, ns[j].uid, s);
							fprintf(tmp, "%4d %4d %e\n", inenzid[i], ns[j].uid, s);

							if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 )
								fprintf(iccof, "%4d %4d %12s %12s %e %e\n", inenzid[i], ns[j].uid, inenzyme[inenzid[i]-1], ns[j].name , s, score);
							//fprintf(tmp, "%4d %4d %4d %e %e\n", i+1, j+1, j+1, s, score);
							cnt++;
						}

						if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 ){
							fprintf(icallf, "%4d %4d %12s %12s %e %e\n", inenzid[i], ns[j].uid, inenzyme[inenzid[i]-1], ns[j].name , s, score);
						}
						//score=exp(log(toccur/(ch*oc))/log(2));
						//fprintf(tmp, "%4d %4d %4d %e\n", i+1, j+1, j+1, score);
					} else {
						if( inenz[i][3] != nsec[j][3] ){
							ch=oc=0;
							for(k=0;k<nsnum;k++){
								if( (inenz[i][0] == nsec[k][0])
								 && (inenz[i][1] == nsec[k][1])
								 && (inenz[i][2] == nsec[k][2]) ){
									ch++;
									oc += occur[k];
								}
							}
							score=log(((double)ch*(double)oc)/((double)toccur))/log(2.0);
							s=exp(score);
							if(score <= CORR_EC_CUTOFF) {
								//fprintf(tmp, "%4d %4d %4d %e\n", inenzid[i], ns[j].uid, ns[j].uid, s);
								fprintf(tmp, "%4d %4d %e\n", inenzid[i], ns[j].uid, s);
								if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 )
									fprintf(iccof, "%4d %4d %12s %12s %e %e\n", inenzid[i], ns[j].uid, inenzyme[inenzid[i]-1], ns[j].name , s, score);
								//fprintf(tmp, "%4d %4d %4d %e %e\n", i+1, j+1, j+1, s, score);
								cnt++;
							}

							if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 ){
								fprintf(icallf, "%4d %4d %12s %12s %e %e\n", inenzid[i], ns[j].uid, inenzyme[inenzid[i]-1], ns[j].name , s, score);
							}
							//score=exp(log(toccur/(ch*oc))/log(2));
							//fprintf(tmp, "%4d %4d %4d %e\n", i+1, j+1, j+1, score);
						} else {
							ch=oc=0;
							for(k=0;k<nsnum;k++){
								if( (inenz[i][0] == nsec[k][0])
								 && (inenz[i][1] == nsec[k][1])
								 && (inenz[i][2] == nsec[k][2])
								 && (inenz[i][3] == nsec[k][3]) ){
									ch++;
									oc += occur[k];
								}
							}
							score=log((double)oc/((double)toccur))/log(2.0);
							s=exp(score);
							if(score <= CORR_EC_CUTOFF) {
								//fprintf(tmp, "%4d %4d %4d %e\n", inenzid[i], ns[j].uid, ns[j].uid, s);
								fprintf(tmp, "%4d %4d %e\n", inenzid[i], ns[j].uid, s);
								if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 )
									fprintf(iccof, "%4d %4d %12s %12s %e %e\n", inenzid[i], ns[j].uid, inenzyme[inenzid[i]-1], ns[j].name , s, score);
								//fprintf(tmp, "%4d %4d %4d %e %e\n", i+1, j+1, j+1, s, score);
								cnt++;
							}

							if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 ){
								fprintf(icallf, "%4d %4d %12s %12s %e %e\n", inenzid[i], ns[j].uid, inenzyme[inenzid[i]-1], ns[j].name , s, score);
							}
							//score=exp(log(toccur/oc)/log(2));
							//fprintf(tmp, "%4d %4d %4d %e\n", i+1, j+1, j+1, score);
						}
					}
				}
			}
		}
	}

	rewind(tmp);

	fprintf(outf, "%d\n", cnt);
	while( fgets(buff1, 4096, tmp) ){
		fprintf(outf, "%s", buff1);
	}

	if( RXNNETWORK_OUTPUT_ALL_INFORMATIONCONTENT_PAIRS == 1 ){
		fclose( icallf );
		fclose( iccof  );
	}

	fclose(outf);
	for(i=0;i<inenzno;i++)
		if( inenz[i] != NULL )
			free(inenz[i]);
	if( inenz != NULL )
		free(inenz);

	for(i=0;i<nsnum;i++)
		if( nsec[i] != NULL )
			free(nsec[i]);
	if( nsec != NULL )
		free(nsec);

	if( occur != NULL )
		free(occur);

	return;
}

//allocate space for graph
void allocate_space(int ** g, int size){
	int i;

	//if( !(g=malloc(size*sizeof(int *))) )
	//	fprintf(stdout, "out of memory!\n"), exit(0);
	for(i=0;i<size;i++)
		if( !(g[i]=malloc(size*sizeof(int))) )
			fprintf(stdout, "out of memory!\n"), exit(0);
}

//free space of graph
void free_space(int ** g, int size){
	int i;
	for(i=0;i<size;i++)
		free(g[i]);
	free(g);
}

//floyd's algorithm
void floyd(int ** g, int ** dist, int ** path, int size){
	int i, j, k;

	//Initialization
	for(i=0; i<size;i++){
		for(j=0;j<size;j++){
			dist[i][j]=g[i][j];
			if( (g[i][j] >= SHRT_MAX) || (i==j) ){
				path[i][j]=-1;
			} else
				path[i][j]=i;
		}
		dist[i][i]=0;
	}

	//The working loop
	//print_dist_matrix(path, size);
	for(k=0;k<size;k++){
		for(i=0;i<size;i++)
			for(j=0;j<size;j++)
				if( (dist[i][k] != SHRT_MAX) && (dist[k][j] != SHRT_MAX)
					&& (dist[i][k]+dist[k][j] < dist[i][j])  ){
					dist[i][j]=dist[i][k]+dist[k][j];
					path[i][j]=path[k][j];
				}
		//fprintf(stdout, "***%d***\n", k);
		//print_dist_matrix(path, size);
	}
}

//test whether there exists a negative cycle in the graph
char test_negative_cycle(int ** dist, int size){
	int i;

	for(i=0;i<size;i++)
		if(dist[i][i]<0)
			return TRUE;

	return FALSE;
}


//recover the shortest path
void output_all_pairs_shortest_path(int ** path, int start, int end) {
//fprintf(stdout, "oapsp:  start=%d end=%d  %d\n", start, end, path[start][end]);
	if( start == end)  {
		//fprintf(stdout, " %2d", start);
		spath[spath_count++]=start;
	} else if (path[start][end] == -1){
		//fprintf(stdout, " No path from %2d to %2d!\n", start, end);
		spath[spath_count]=-1;
	} else {
		output_all_pairs_shortest_path(path, start, path[start][end]);
		//fprintf(stdout, " %2d", end);
		spath[spath_count++]=end;
	}
}

//print the distance matrix
void print_dist_matrix(int ** dist, int size, char * fname){
	FILE *outf;//, *outf1;
	int i, j;
	int * count, maxlen, totcnt, totnoway, cum;

	if( !(outf=fopen(fname, "w")) )
		fprintf(stdout, "%s cannot be opened for writing!\n", fname), exit(0);

	//fprintf(outf, "      "); 	//for(i=0;i<size;i++) 	//	fprintf(outf, " %5d", i+1);
	for(i=0;i<size;i++){
		fprintf(outf, "\n %1d", i+1);
		for(j=0;j<size;j++){
			//fprintf(outf, " %1d", dist[i][j]);
			if(dist[i][j] == SHRT_MAX){
				fprintf(outf, " /");
			} else
				fprintf(outf, " %1d", dist[i][j]);
		}
	}
	//fprintf(stdout, "\n\n");
	fclose(outf);

	//statistics
	//if( !(outf1=fopen("shrtpath_stat", "w")) )
	//	fprintf(stdout, "%s cannot be opened for writing!\n", fname), exit(0);

	//find the longest shortest path
	//fprintf(outf1, "Number of vertices: %d\n\n", size);
	maxlen=0;
	for(i=0;i<size;i++){
		for(j=0;j<size;j++){
			if( dist[i][j] != SHRT_MAX && dist[i][j] > maxlen)
				maxlen= dist[i][j];
		}
	}

	if( !(count = malloc( (maxlen +1) * sizeof(int))) )
		fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);
	for(i=0;i<=maxlen; i++)
		count[i] = 0;

	totnoway = 0;
	for(i=0;i<size;i++){
		for(j=0;j<size;j++){
			if( dist[i][j] != SHRT_MAX ){
				(count[ dist[i][j] ])++;
			} else {
				totnoway++;
			}
		}
	}

	totcnt = 0;
	for(i=1; i<=maxlen; i++)
		totcnt += count[i];

	cum = 0;
	for(i=0; i<=maxlen; i++){
		cum += count[i];
		//fprintf(outf1, "%2d: %10d -- %5.2f%%  %5.2f%%\n", i, count[i], (double)count[i]*100/((double)totcnt), (double)cum*100/((double)totcnt));
	}

	//fprintf(outf1, "\n\nNumber of pairs without pathway: %d", totnoway);

	//fclose(outf1);

	if( count != NULL )
		free(count);

	return;
}

//print the path matrix
void print_path_matrix(int ** dist, int size, char * fname){
	FILE * outf;
	int i, j;

	if( !(outf=fopen(fname, "w")) )
		fprintf(stdout, "%s cannot be opened for writing!\n", fname), exit(0);

	for(i=0;i<size;i++){
		fprintf(outf, "\n %1d", i+1);
		for(j=0;j<size;j++){
			fprintf(outf, " %1d", dist[i][j]+1);
		}
	}
	fclose(outf);

	return;
}

//read in the distance matrix
void read_dist_matrix(int ** dist, int size, char * fname){
	FILE * inf;
	int i, j;
	long len;
	char * buff;
	char * token;

	len=(size+1)*10;

	if( !(buff=malloc(len*sizeof(char))) )
		fprintf(stdout, "out of memory!\n"), exit(0);

	if( !(inf=fopen(fname, "r")) )
		fprintf(stdout, "%s cannot be opened for reading!\n", fname), exit(0);

	//fprintf(outf, "      "); 	//for(i=0;i<size;i++) 	//	fprintf(outf, " %5d", i+1);
	i=0;
	while( fgets(buff, len, inf) ){
		if(strlen(buff) > 20){
			if( (i+1) > size)
				fprintf(stdout, "(more) data error in the file %s\n", fname), exit(0);

			token=strtok(buff, " ");
			i=atoi(token)-1;
			j=0;
			while( (token=strtok(NULL, " ")) ){
				if( (j+1) > size)
					fprintf(stdout, "(more) data error in the line %d (in the file %s)\n", i+1, fname), exit(0);
				if( strstr(token, "/") ){
					dist[i][j++]=SHRT_MAX;
				} else
					dist[i][j++]=atoi(token);
			}
			if(j != size)
				fprintf(stdout, "(less) data error in the line %d (in the file %s)\n", i+1, fname), exit(0);
		}
	}

	if( (i+1) != size)
		fprintf(stdout, "(less) data error in the file %s\n", fname), exit(0);

	fclose(inf);
	free(buff);

	return;
}

//read in the path matrix
void read_path_matrix(int ** dist, int size, char * fname){
	FILE * inf;
	int i, j;
	long len;
	char * buff;
	char * token;

	len=(size+1)*10;

	if( !(buff=malloc(len*sizeof(char))) )
		fprintf(stdout, "out of memory!\n"), exit(0);

	if( !(inf=fopen(fname, "r")) )
		fprintf(stdout, "%s cannot be opened for reading!\n", fname), exit(0);

	//fprintf(outf, "      ");
	//for(i=0;i<size;i++)
	//	fprintf(outf, " %5d", i+1);
	i=0;
	while( fgets(buff, len, inf) ){
		if(strlen(buff) > 20){
			if( (i+1) > size)
				fprintf(stdout, "(more) data error in the file %s\n", fname), exit(0);

			token=strtok(buff, " ");
			i=atoi(token)-1;
			j=0;
			while( (token=strtok(NULL, " ")) ){
				if( (j+1) > size)
					fprintf(stdout, "(more) data error in the line %d (in the file %s)\n", i+1, fname), exit(0);
				dist[i][j++]=atoi(token)-1;
			}
			if(j != size)
				fprintf(stdout, "(less) data error in the line %d (in the file %s)\n", i+1, fname), exit(0);
		}
	}

	if( (i+1) != size)
		fprintf(stdout, "(less) data error in the file %s\n", fname), exit(0);

	fclose(inf);
	free(buff);

	return;
}


//allocate space for corr
void alloc_space(pair * co, int num){
	if( !(co = (pair *)malloc(num*sizeof(pair))) ){
		fprintf(stdout, "Out of memory!\n");
		exit(0);
	}
}

//read in correspondence information
void correspondence_reader(pair ** corr, char * fname, int * num, int n, char **names, char **qnames, int q_size){
	FILE * inf;
	char buffer[4096], str[200], str2[200];
	int i, j, k;
	double d;
	pair * co;

	if( !(inf=fopen(fname, "r")) )
		fprintf(stdout, "The file %s cannot be opened for reading!\n", fname), exit(0);

	i=0;
	while( fgets(buffer, 4096, inf) ) {
		if( ! is_blank(buffer) )
			++ i;
	}
	*num = i;
	if( !( *corr = co = calloc(i, sizeof(pair))) )
		fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);

	rewind(inf);

	k = 0;
	while( fgets(buffer, 4096, inf) ){
		if( ! is_blank(buffer) ){		
			if (sscanf(buffer, " %s %s %le ", str, str2, &d) == 3){
				for( i=0; i<q_size; ++i ){
					if( ! strcmp(qnames[i], str) )
						break;
				}
				if( i == q_size )
					fprintf(stdout, "\ncorr: i == q_size\n"), fflush(NULL), exit(0);

				for( j=0; j<n; ++j ){
					if( ! strcmp(names[j], str2) )
						break;
				}
				if( j == n ){
					fprintf(stdout, "\nstr2: %s\n", str2);
					fprintf(stdout, "\ncorr: j == n\n"), fflush(NULL), exit(0);
				}

				co[k].v1 = i;
				co[k].v2 = j;
				co[k].sim = d;
			}  else {
				fprintf(stdout, "*3* Missing data in the file %s.\n", fname);
				fflush(NULL), exit(0);
			}
			
			k++;
		}
	}
	fclose(inf);
	return;
}

//generate the correspondence graph. vertices--correspondence pairs; edge -- length of shortest paths
void generate_corr_graph(int ** dist1, int ** dist2, pair * co, int ** g, int n1, int n2, int n3){
	int i, j;
	int len1, len2;

	for(i=0;i<n3;i++){
		g[i][i]=0;
		for(j=i+1;j<n3;j++){
			len1=dist1[co[i].v1][co[j].v1];
			len2=dist2[co[i].v2][co[j].v2];
			g[i][j]=g[j][i]=(len1>=len2)?(len1):(len2);
		}
	}
}

//<********* path searching ********<

//find k shortest pathways with top scores using Eppstein's algorithm. 
void pathway_query_top_paths3(double ** dag, int inplen, int ** inpco, double ** inpcowt, int *inpcosize, int cdorder,
							 int * cdlorder, int ** pth, int nno, node *no, char **inpname, int *inpid,
							 pair * co, int cono, int flag, char * outfname){
	FILE * outf, *gfile, *inf;
	char cmdline[100], buffer[4096], *ptr1, *ptr;
	int i,j, nnum, x, y,z, m, n, ednum, topcnt, xxa;
	pathvert * toppaths;

	nnum=cdorder+2;

	if( !(outf=fopen(outfname, "w")) )
		fprintf(stdout, "The file to store the results of top pathway queries cannot be opened for writing!\n"),fflush(stdout), exit(0);

	//fprintf(outf, "\nPathway Query Input:  ");
	//for(i=0;i<inplen;i++)
	//	fprintf(outf, "  %1s", inpname[i]);
	//fflush(outf);

	//output graph file in correct format
	if( !(gfile = fopen("tmpgraph", "w")) )
		fprintf(stdout, "tmpgraph cannot be opened for writing!\n"), fflush(NULL), exit(0);
	fprintf(gfile, "n %d", nnum);
	//count the number of edges
	ednum = m=0;
	for( i=0; i<=inplen-2; i++ ){
		for(j=0;j<inpcosize[i];j++){
			n=0;
			for(x=i+1;x<=inplen-1;x++){
				for(y=0;y<inpcosize[x];y++){
					if( n>=cdlorder[i] )
						fprintf(stdout, "out of boundary!\n"), fflush(stdout), exit(0);

					if( dag[m][n] < FLT_MAX )
						ednum++;

					n++;
				}
			}
			if( n != cdlorder[i] )
				fprintf(stdout, "Less members!\n"), fflush(stdout), exit(0);

			m++;
		}
	}
	//edges from source and to sink
	if(GENERAL == 1){
		ednum += cdorder*2;
	} else {
		ednum += (inpcosize[0] + inpcosize[inplen-1]);
	}
	fprintf(gfile, "\nm %d", ednum);
	fprintf(gfile, "\ns 1");
	fprintf(gfile, "\nt %d", nnum);

	//output edges from the source
	if(GENERAL == 1){
		m=2;
		for(i=0; i<inplen; i++){
			for(j=0; j<inpcosize[i]; j++){
				fprintf(gfile, "\na %d %d %le", 1, m, inpcowt[i][j]+INDEL_PENALTY*i);				
				m++;
			}
		}
	} else {
		for(j=0;j<inpcosize[0];j++)
			fprintf(gfile, "\na %d %d %le", 1, j+2, inpcowt[0][j]);		
	}
	//output edges to the sink
	if(GENERAL == 1){
		m=2;
		for(i=0; i<inplen; i++){
			for(j=0; j<inpcosize[i]; j++){
				fprintf(gfile, "\na %d %d %le", m, nnum, INDEL_PENALTY*(inplen-1-i) );				
				m++;			
			}
		}
	} else {		
		for(j=0;j<inpcosize[inplen-1];j++){
			fprintf(gfile, "\na %d %d %le", nnum-1-j, nnum, 0.0);			
		}
	}
	//output edges between correspondences
	m=0;
	for( i=0; i<=inplen-2; i++ ){
		for(j=0;j<inpcosize[i];j++){
			n=0;
			z = nnum - cdlorder[i];
			for(x=i+1;x<=inplen-1;x++){
				for(y=0;y<inpcosize[x];y++){
					if( n>=cdlorder[i] )
						fprintf(stdout, "out of boundary!\n"), fflush(stdout), exit(0);

					if( dag[m][n] < FLT_MAX )
						fprintf(gfile, "\na %d %d %le", m+2, z+n, dag[m][n]);	
					
					n++;
				}
			}
			if( n != cdlorder[i] )
				fprintf(stdout, "Less members!\n"), fflush(stdout), exit(0);

			m++;
		}
	}
	fprintf(gfile, "\n");
	fclose(gfile);

	sprintf(cmdline, "eppstein_modified tmpgraph %d -paths", NO_OF_TOP_PATHS_OUTPUT);

	//calculate the pathways with top scores
	//call Eppstein's algorithm
	system(cmdline); 

	//read in eppstein results
	if( !(inf = fopen("tmpeppout", "r")) ){
		//fprintf(stdout, "File tmpeppout cannot be opened for reading!\n"), fflush(NULL), exit(0);
		fflush(NULL); 
		return;
	}

	topcnt = 0;
	while( fgets(buffer, 4096, inf) ){
		if( strstr(buffer, "N=") )
			topcnt++;
	}
	rewind(inf);

	if( !(toppaths=malloc(topcnt * sizeof(pathvert))) )
		fprintf(stdout, "out of memory!\n"), fflush(stdout), exit(0);
	
	i=0;
	while( fgets(buffer, 4096, inf) ){
		if( strstr(buffer, "N=") ){
			if( ptr = strstr(buffer, "(Cost:") ){
				if( sscanf(ptr, "(Cost: %le) ", &(toppaths[i].score)) != 1 )
					fprintf(stdout, "\nCan't correctly read cost in line:\n%s\n\n", buffer), fflush(NULL), exit(0);

				while( (ptr - buffer) >= 0 && *ptr != '-' ) ptr --;
				ptr --;
				while( (ptr - buffer) >= 0 && *ptr != '-' ) ptr --;
				*(ptr+1) = '\0';

				ptr1 = buffer;
				while( *ptr1 != '-' )
					ptr1++;
				ptr1++;

				if( ptr1 > ptr ){
					toppaths[i].len = 0;
					i++;
					continue;
				}

				j=0;
				ptr = ptr1;
				while( *ptr != '\0' ){
					if( *ptr == '-' )
						j++;
					ptr++;
				}

				toppaths[i].len = j;
				if( !(toppaths[i].path = malloc( toppaths[i].len *sizeof(int) )) )
					fprintf(stdout, "out of memory for toppaths[i].len\n"), fflush(NULL), exit(0);

				j=0;
				while( *ptr1 != '\0' ){
					ptr = ptr1;
					while( *ptr != '-' ) ptr++;
					*ptr = '\0';
					(toppaths[i].path)[j] = atoi( ptr1 ) - 1;
					j++;
					ptr1 = ptr + 1;
				}

				i++;
			} else {
				fprintf(stdout, "\nNo cost in this line:\n%s\n\n", buffer), fflush(NULL),exit(0);
			}

		}		
	}
	fclose(inf);

	//output parameters
	//for(i=0; i<ARGC; ++i)
	//	fprintf(outf, "%s ", ARGV[i]);
	//fprintf(outf, "\n"), fflush(NULL);

	fprintf(outf, "%s ", ARGV[0]);
	fprintf(outf, "-q=%1s ", seqoutfname1);
	fprintf(outf, "-i=%1s ", network_file);
	fprintf(outf, "-c=%1s ", corrfilename);
	fprintf(outf, "-g=%1d ", MAX_NO_OF_MISMATCHES_AND_GAPS);
	fprintf(outf, "-l=%1d ", if_blast2seq_pathwayquery);
	fprintf(outf, "-s=%1d ", GENERAL);
	fprintf(outf, "-p=%3.1f ", INDEL_PENALTY);
	fprintf(outf, "-n=%1d ", NO_OF_TOP_PATHS_OUTPUT);
	fprintf(outf, "-o=%1s ", topoutfname);
	//fprintf(outf, "\n"), fflush(NULL);

	//output top paths
	for(x=0; x<topcnt; x++){
		//reconstruct a path between two nodes
		spath_count=(toppaths[x].len) + 2;
		spath[0]=0;
		for(i=0;i<(toppaths[x].len);i++)
			spath[i+1]=(toppaths[x].path)[i];
		spath[i+1]=nnum-1;

		if(spath_count == 0){
			fprintf(outf, "\n No similar pathways found!\n\n");
		} else {
			pathway_query_recover_path(inplen, inpco, inpcosize, pth, nno, no, inpname, inpid);
			//begin here

			fprintf(outf, "\n\nResult%02d: score=%3.2f  ", x+1, -toppaths[x].score);

			for(i=0;i<finalpath_count2;i++){
				if( finalpath1[i] != -1 ){
					fprintf(outf, "\n %-13s", inpname[finalpath1[i]]);
				} else {
					fprintf(outf, "\n   -          ");
				}

				if( finalpath2[i]==-1 ){
					fprintf(outf, "  ");
					fprintf(outf, "    -      ");
				} else {
					if( finalpath1[i] != -1 ){
						//fprintf(outf, " %-13s", inpname[finalpath1[i]]);
						for( xxa=0; xxa < cono; ++ xxa){
							if( co[xxa].v1 == finalpath1[i] && co[xxa].v2 == finalpath2[i] )
								break;
						}

						if( xxa < cono ){
							fprintf(outf, "--");
							if ( no[finalpath2[i]].name ){
								fprintf(outf, " %-13s", no[finalpath2[i]].name);
							} else {
								fprintf(outf, "    X      ");
							}								
						} else {
							fprintf(outf, "  ");
							if ( no[finalpath2[i]].name ){
								fprintf(outf, " %-13s", no[finalpath2[i]].name);
							} else {
								fprintf(outf, "    X      ");
							}										
						}
					} else {
						fprintf(outf, "  ");
						if ( no[finalpath2[i]].name ){
							fprintf(outf, " %-13s", no[finalpath2[i]].name);
						} else {
							fprintf(outf, "    X      ");
						}						
					}
				}
			}

		/*
			fprintf(outf, "\nquery:");
			for(i=0;i<finalpath_count1; i++){
				if( finalpath1[i] != -1 ){
					fprintf(outf, " %-13s", inpname[finalpath1[i]]);
				} else {
					fprintf(outf, "      -       ");
				}
			}
			
			fprintf(outf, "\n path:");
			for(i=0;i<finalpath_count2;i++){
				if( finalpath2[i]==-1 ){
					fprintf(outf, "       -      ");
				} else {
					if( finalpath1[i] != -1 ){
						//fprintf(outf, " %-13s", inpname[finalpath1[i]]);
						for( xxa=0; xxa < cono; ++ xxa){
							if( co[xxa].v1 == finalpath1[i] && co[xxa].v2 == finalpath2[i] )
								break;
						}

						if( xxa < cono ){
							if ( no[finalpath2[i]].name ){
								fprintf(outf, " %-13s", no[finalpath2[i]].name);
							} else {
								fprintf(outf, "       X      ");
							}								
						} else {
							if ( no[finalpath2[i]].name ){
								fprintf(outf, " *%-12s", no[finalpath2[i]].name);
							} else {
								fprintf(outf, "      *X      ");
							}										
						}
					} else {
						if ( no[finalpath2[i]].name ){
							fprintf(outf, " %-13s", no[finalpath2[i]].name);
						} else {
							fprintf(outf, "       X      ");
						}						
					}
				}
			}
		*/
			
			//fprintf(outf, "\n\n");
		}
	}
	fprintf(outf, "\n");

	fclose(outf);
}

//show commandline parameters
void infobox(){
	fprintf(stdout, "\npathmatch ");

	//NETWORK_TYPE
	//fprintf(stdout, "\n\t-t=  type of network to be created");
	//fprintf(stdout, "\n\t\t0  -- protein interaction network");
	//fprintf(stdout, "\n\t\t1  -- reaction network");

	//input query file name
	fprintf(stdout, "\n\t-q=  file name containing query path");

	//input network file name
	fprintf(stdout, "\n\t-i=  file name containing input graph");

	//correspondence file name
	fprintf(stdout, "\n\t-c=  file name containing correspondences between vertices");

	// maximum number of mismatches and gaps allowed -- MAX_NO_OF_MISMATCHES_AND_GAPS=2;
	fprintf(stdout, "\n\t-g=  maximum number of mismatches or indels between two matches ");

	//if_blast2seq_pathwayquery
	fprintf(stdout, "\n\t-l=  type of match score");
	fprintf(stdout, "\n\t\t0  -- use similarity score as match score");
	fprintf(stdout, "\n\t\t1  -- use negative logarithm of similarity score as match score");

	//int GENERAL=1;
	fprintf(stdout, "\n\t-s=  whether gaps are allowed at the start and the end of path alignment");
	fprintf(stdout, "\n\t\t0  -- not allowed, i.e., must be matches at the start and the end");//, i.e. the first and last nodes of the input pathway must have correspondences");
	fprintf(stdout, "\n\t\t1  -- allowed");

	//indel penalty
	//INDEL_PENALTY=-log(CORR_E_VALUE_CUTOFF)*5;
	fprintf(stdout, "\n\t-p=  mismatch and indel penalty");

	//int NO_OF_TOP_PATHS_OUTPUT 60
	fprintf(stdout, "\n\t-n=  number of output paths");

	//output file name
	fprintf(stdout, "\n\t-o=  output file name");

	//cutoffs for recording bl2seq results
	//double E_VALUE_CUTOFF  = 1.0; //1e-2
	//fprintf(stdout, "\n\t-b=  cutoff value for recording bl2seq results");
	//int    E_LENGTH_CUTOFF =  0;

	//cutoff of similarity scores for determining correspondence
	//double CORR_E_VALUE_CUTOFF= 1; //1e-5;
	//fprintf(stdout, "\n\t-c=  cutoff of similarity scores for determining correspondence");
	//int CORR_E_LENGTH_CUTOFF  = 0;

	//cutoffs for determining similarity according to EC
	//double CORR_EC_CUTOFF=-1.7;
	//fprintf(stdout, "\n\t-e=  cutoff of similarity scores according to EC");

	//if_de_novo_graph1
	//fprintf(stdout, "\n\t-v=  1 apply all-pair shortest paths algorithm; 0 otherwise");

	//cpd correspondence file name
	//fprintf(stdout, "\n\t-m=  compound correspondence file name");


	fprintf(stdout, "\n\n");
}

//read graphs in linked list representation and save to adjacent matrix
void graph_reader_1(char *f, int *** g, int gsize, char ** names){
	FILE * inf;
	int ** h;
	int  i, j;//hsize,
	char *buffer;
	char *token;
	int max, n;
	char ch;

	if( !(inf = fopen(f, "r")) )
		fprintf(stdout, "File %s can't be opened for reading!\n", f), fflush(NULL), exit(0);

	//allocate space for the graph
	if( !(h = (int **)calloc(gsize, sizeof(int *))) )
		fprintf(stdout, "out of memory for h -- %s\n", f), fflush(NULL), exit(0);
	for(i=0; i<gsize; i++)
		if( !(h[i] = (int *)calloc(gsize, sizeof(int))) )
			fprintf(stdout, "out of memory for h[i] -- %s\n", f), fflush(NULL), exit(0);

	for(i=0; i<gsize; i++){
		for(j=0;j<gsize; j++)
			h[i][j] = SHRT_MAX;
	}

	//determine the max line length
	max = 0; 
	n = 0;
	while( (ch = fgetc(inf)) != EOF ){
		if( ch == '\n' ){
			++n;
			if( n > max )
				max = n;
			n  = 0;
		} else 
			++ n;		
	}
	if( n > max )
		max = n;
	max += 2;

	if( !(buffer = malloc(max * sizeof(char))) )
		fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

	rewind(inf);

	while( fgets(buffer, max, inf) ){
		if( !is_blank(buffer) ){
			if( (token=strtok(buffer, ", \n")) ) {
				for(i=0; i<gsize; ++i){
					if( ! strcmp(names[i], token) )
						break;
				}

				if( i == gsize )
					fprintf(stdout, "\ni == gsize\n"), fflush(NULL), exit(0);				
			}

			while( (token=strtok(NULL, ", \n")) ) {
				for(j=0; j<gsize; ++j){
					if( ! strcmp(names[j], token) )
						break;
				}

				if( j == gsize ){
					fprintf(stdout, "\n%s\n", token);
					fprintf(stdout, "\nj == gsize\n"), fflush(NULL), exit(0);	
				}

				if( j >= 0 )
					h[i][j] = 1;
			}	
		}
	}

	*g = h;

	fclose( inf );

	for(i=0; i<gsize; i++){
		for(j=i+1;j<gsize; j++){
			if( h[i][j] != h[j][i] ){
				if_directed_graph = 1;
				return;
			}
		}
	}

	return;
}

//given a graph, return number of edges
int count_graph_edge(int ** g, int n){
	int i,j, m=0;

	for(i=0; i<n; i++)
		for(j=i+1; j<n; j++)
			if( g[i][j] == 1 )
				m++;

	return m;
}

typedef struct strlist {
	char * s;
	struct strlist * next;
} strlist;

//read in node names
void nname_reader(char *fname, int * size, char ***nnames){
	FILE * inf;
	char *buff, *buffer, ** s;
	int n, max;
	char tmpstr[200], ch;
	char *ptr;

	strlist * tail, *curr;
	strlist * slist = NULL;

	if( !(inf = fopen(fname, "r")) ){
		*nnames = NULL;
		return;
	} 

	//determine the max line length
	max = 0; 
	n = 0;
	while( (ch = fgetc(inf)) != EOF ){
		if( ch == '\n' ){
			++n;
			if( n > max )
				max = n;
			n  = 0;
		} else 
			++ n;		
	}
	if( n > max )
		max = n;
	max += 2;

	if( !(buff = malloc(max * sizeof(char))) )
		fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);
	if( !(buffer = malloc(max * sizeof(char))) )
		fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

	rewind(inf);

	//read in the first column
	while( fgets(buff, max, inf) ){
		if( ! is_blank( buff ) ){
			if( ! slist ){
				if( !( slist = calloc(1, sizeof(strlist))) )
					fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

				if( sscanf(buff, " %s %*s ", tmpstr) != 1 )
					fprintf(stderr, "\nwrong here\n%s", buff), fflush(NULL), exit(0);

				if( !( slist->s = calloc(strlen(tmpstr)+1, sizeof(char))) )
					fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

				strcpy(slist->s, tmpstr);
				slist->next = NULL;

				tail = slist;
			} else {
				if( !( curr = calloc(1, sizeof(strlist))) )
					fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

				if( sscanf(buff, " %s %*s ", tmpstr) != 1 )
					fprintf(stderr, "\nwrong here\n%s", buff), fflush(NULL), exit(0);

				if( !( curr->s = calloc(strlen(tmpstr)+1, sizeof(char))) )
					fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

				strcpy(curr->s, tmpstr);
				curr->next = NULL;

				tail->next = curr;
				tail = tail->next;
			}			
		}
	}

	rewind(inf);

	//read other names, if any
	while( fgets(buff, max, inf) ){
		if( ! is_blank( buff ) ){
			strcpy(buffer, buff);

			ptr = strtok(buffer, ", \n");

			while( (ptr=strtok(NULL, ", \n")) ) {				
				strcpy(tmpstr, ptr);

				curr = slist;
				while( curr ){
					if( ! strcmp(curr->s, tmpstr) ){
						break;
					}

					curr = curr->next;
				}

				if( ! curr ){
					if( !( curr = calloc(1, sizeof(strlist))) )
						fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

					if( !( curr->s = calloc(strlen(tmpstr)+1, sizeof(char))) )
						fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

					strcpy(curr->s, tmpstr);
					curr->next = NULL;

					tail->next = curr;
					tail = tail->next;
				}				
			}			
		}
	}

	//count # of nodes
	n = 0;
	curr = slist;
	while( curr ){
		++ n;
		curr = curr->next;
	}

	*size = n;

	if( !(s = calloc(n, sizeof(char *))) )
		fprintf(stderr, "\nout of memory!\n"), fflush(NULL), exit(0);

	n = 0;
	curr = slist;
	while( curr ){
		if( !(s[n] = calloc(strlen(curr->s)+1, sizeof(char))) )
			fprintf(stderr, "\nout of memory!\n"), fflush(NULL), exit(0);

		strcpy(s[n], curr->s);
		
		++ n;
		curr = curr->next;
	}
	
	fclose(inf);

	*nnames = s;

	return;
}


